import ctypes
from distutils.version import LooseVersion
import io
import os
import platform
import re

from anaconda_navigator.api.download_api import DownloadAPI
from anaconda_navigator.config import (BITS_32, BITS_64, CONF_PATH, HOME_PATH,
                                       LINUX, MAC, OS_64_BIT, WIN)
from anaconda_navigator.utils.logs import logger


class BaseApp(object):
    def __init__(self, config, filename,
                 windows_ext, windows_folder_name,
                 mac_name, display_name, app_name,
                 description, image_path, needs_license,
                 non_conda, distro_map, process_api, conda_api):
        self.config = config
        self.filename = filename
        self.windows_ext = windows_ext
        self.windows_folder_name = windows_folder_name
        self.mac_name = mac_name
        self.display_name = display_name
        self.app_name = app_name
        self.description = description
        self.image_path = image_path
        self.needs_license = needs_license
        self.non_conda = non_conda
        self.distro_map = distro_map
        self._process_api = process_api
        self._conda_api = conda_api

        # These download methods return a worker
        self._download_api = DownloadAPI(config=config)
        _is_valid_url = self._download_api.is_valid_api_url
        self.download = self._download_api.download
        self.download_is_valid_api_url = _is_valid_url
        self.download_is_valid_channel = self._download_api.is_valid_channel
        self.download_terminate = self._download_api.terminate

        self.init()

    # --- App Handling
    # -------------------------------------------------------------------------
    def init(self):
        """Initialize App setttings for install."""
        DISTRO_NAME = ''
        DISTRO_VER = ''
        INST_DIR_FOUND = None

        self._DISTRO_VER = None

        if WIN:
            SUBDIR, INST_EXT, INST_DIR = self._find_win_install_dir()
        elif MAC:
            SUBDIR, INST_EXT, INST_DIR = self._find_mac_install_dir()
        elif LINUX:
            SUBDIR, INST_EXT, INST_DIR = self._find_linux_install_dir()

        INSTFILE = os.path.join(
            CONF_PATH,
            'temp',
            'apptemp.{}'.format(INST_EXT),
        )

        self._DISTRO_NAME = DISTRO_NAME
        self._DISTRO_VER = DISTRO_VER
        self._SUBDIR = SUBDIR
        self._INST_DIR = INST_DIR
        self._INSTFILE = INSTFILE

    def _find_win_install_dir(self):
        from anaconda_navigator.external.knownfolders import (
            get_folder_path,
            FOLDERID,
        )
        INST_EXT = 'exe'

        _kernel32 = ctypes.windll.kernel32
        _windir = ctypes.create_unicode_buffer(1024)
        _kernel32.GetWindowsDirectoryW(_windir, 1024)
        _windrive = _windir.value[:3]

        if BITS_32:
            SUBDIR = 'win32-user'
            PROGRAM_FILES = get_folder_path(FOLDERID.ProgramFilesX86)[0]
            _fallback = os.path.join(_windrive, 'Program Files (x86)')
        elif BITS_64:
            SUBDIR = 'win32-x64-user'
            PROGRAM_FILES = get_folder_path(FOLDERID.ProgramFilesX64)[0]
            _fallback = os.path.join(_windrive, 'Program Files')

        LOCAL_DATA = get_folder_path(FOLDERID.LocalAppData)[0]
        PROGRAM_FILES_64 = (
            get_folder_path(FOLDERID.ProgramFilesX64)[0]
            or os.path.join(_windrive, 'Program Files')
        )

        if PROGRAM_FILES is None:
            PROGRAM_FILES = os.environ.get('ProgramFiles', _fallback)

        # Check the correct location System vs. User bit independent
        INST_DIR_FOUND = None
        if os.path.exists(
            os.path.join(PROGRAM_FILES, self.windows_folder_name)
        ):
            INST_DIR_FOUND = os.path.join(
                PROGRAM_FILES, self.windows_folder_name
            )
        elif PROGRAM_FILES_64 and os.path.exists(
            os.path.join(PROGRAM_FILES_64, self.windows_folder_name)
        ):
            INST_DIR_FOUND = os.path.join(
                PROGRAM_FILES_64, self.windows_folder_name
            )
        else:
            INST_DIR_FOUND = os.path.join(
                LOCAL_DATA, self.windows_folder_name
            )
            if not os.path.isdir(INST_DIR_FOUND):
                INST_DIR_FOUND = os.path.join(LOCAL_DATA, 'Programs', self.windows_folder_name)

        return SUBDIR, INST_EXT, INST_DIR_FOUND

    def _find_mac_install_dir(self):
        SUBDIR = 'darwin'
        INST_EXT = 'zip'
        INST_DIR = os.path.join('/Applications')
        return SUBDIR, INST_EXT, INST_DIR

    def _find_linux_install_dir(self):
        raise NotImplementedError

    @property
    def executable(self):
        quote = '"' if ' ' in self._EXE else ''
        return quote + self._EXE + quote if os.path.exists(self._EXE) else None

    @property
    def is_available(self):
        """Is App available for installation on this platform."""
        try:
            if WIN:
                # Not checking XP, let their installer cry
                return True
            elif MAC:
                v = LooseVersion(platform.mac_ver()[0]) >= LooseVersion('10.11')
                return v
            elif LINUX:
                _distro_min_ver = self._DISTRO_MAP[self._DISTRO_NAME]
                return (
                    (len(self._DISTRO_NAME) > 0)
                    and (len(self._DISTRO_VER) > 0) and (
                        LooseVersion(self._DISTRO_VER) >=
                        LooseVersion(_distro_min_ver)
                    )
                )  # NOQA
        except Exception:
            return False
        return False

    @property
    def installed(self):
        """Return the installed status of the package."""
        return bool(self.executable)

    def log_path(self, uninstall=False, delete=False):
        """Return the log path for installer/uninstaller."""
        if uninstall:
            fname = '{}-uninstall-log.txt'.format(self.app_name)
        else:
            fname = '{}-install-log.txt'.format(self.app_name)

        log_path = os.path.join(CONF_PATH, 'temp', fname)

        if delete and os.path.isfile(log_path):
            try:
                os.remove(log_path)
            except Exception:
                pass

        return log_path

    def log_data(self, uninstall=False):
        """Return the parsed log data from installer/uninstaller."""
        log_path = self.log_path(uninstall=uninstall)

        if os.path.isfile(log_path):
            with io.open(log_path, 'r') as f:
                f.read()

        # TODO: load data!
        log_data = {'successful': True}
        return log_data

    @property
    def install_enabled(self):
        # Add App global app (if it does not exist as a conda package)
        return self.config.get('home', '{}_enable'.format(self.app_name))

    def create_config_backup(self, data):
        """
        Create a backup copy of the app configuration file `data`.

        Leave only the last 10 backups.
        """
        raise NotImplementedError

    def update_config(self, prefix):
        """Update app python interpreter user config."""
        raise NotImplementedError

    def install(self, password=None):
        """Install app."""
        raise NotImplementedError

    def install_extensions(self):
        """Install app extensions."""
        raise NotImplementedError

    def send_telemtry(self):
        """Send app telemetry."""
        raise NotImplementedError

    def win_uninstaller(self):
        """Check the right uninstaller file on windows."""
        raise NotImplementedError

    def remove(self, password=None):
        """Remove app files from computer or run uninstaller."""
        raise NotImplementedError



import datetime
import errno
import json
import io
import os
import re
import shutil
import sys


from .base import BaseApp

from anaconda_navigator.api.process import DummyWorker
from anaconda_navigator.config import (CONF_PATH, LINUX, LINUX_DEB, HOME_PATH,
                                       LINUX_DNF, LINUX_RPM, MAC, WIN, OS_64_BIT)
from anaconda_navigator.static import scripts, images
from anaconda_navigator.utils.py3compat import PY3, to_binary_string
from anaconda_navigator.utils.logs import logger


class VSCodeApp(BaseApp):
    def __init__(self, **kwargs):
        # minimum supported linux distro versions
        distro_map = {
            'rhel': '7',
            'sles': '12',
            'centos': '7',
            'debian': '8',
            'fedora': '23',
            'suse': '42.1',
            'ubuntu': '14.04'
        }

        super(VSCodeApp, self).__init__(
            app_name='vscode',
            filename='code',
            windows_ext='.cmd',
            windows_folder_name='Microsoft VS Code',
            mac_name='Visual Studio Code.app',
            display_name='VS Code',
            description=('Streamlined code editor with support for '
                         'development operations like debugging, task '
                         'running and version control.'),
            image_path=images.VSCODE_ICON_1024_PATH,
            needs_license=False,
            non_conda=True,
            distro_map=distro_map,
            **kwargs
        )

        self.rpm_asc_file_url = 'https://packages.microsoft.com/keys/microsoft.asc'
        self.app_data = self._application_data()
        if MAC:
            self._EXE = os.path.join(
                    self._INST_DIR,
                    self.mac_name,
                    'Contents/Resources/app/bin/code'
                )
        elif LINUX:
            self._EXE = os.path.join(self._INST_DIR, 'bin', 'code')
        else:
            self._EXE = os.path.join(
                self._INST_DIR, 'bin', 'code.cmd'
            )


    @property
    def versions(self):
        return [self.app_data.get('productVersion')]

    def _version(self):
        """Query the vscode version for the default installation path."""
        # TODO: generalize for external app, not just vscode
        version = None
        if self._vscode_version_value is None:
            if self.executable:
                cmd = [self.executable, '--version']

                import subprocess
                stdout = ''
                stderr = ''
                try:
                    p = subprocess.Popen(
                        cmd,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        executable=self.executable,
                    )
                    stdout, stderr = p.communicate()
                    if PY3:
                        stdout = stdout.decode()
                        stderr = stderr.decode()
                except OSError:
                    pass

                if stdout:
                    output = [o for o in stdout.split('\n') if o and '.' in o]
                    version = output[0]

                self._vscode_version_value = version
        else:
            version = self._vscode_version_value

        return version

    @property
    def version(self):
        """Return the current installed version or the highest version."""
        return self._vscode_version() or self.versions[-1]

    def install_extensions(self):
        """Install app extensions."""
        wm = self._process_api
        logger.debug('Installing vscode extensions')
        cmd = [
            '"{}"'.format(self._EXE),  # must wrap in quotes so that spaces don't mess things up
            '--install-extension',
            'ms-python.anaconda-extension-pack',
            # ms-python-anaconda-extension
            # ms-python.python
        ]
        logger.debug(' '.join(cmd))
        worker = wm.create_process_worker(cmd)
        return worker

    def send_telemetry(self):
        """Send app telemetry."""
        wm = self._process_api
        logger.debug('Sending vscode telemetry')
        cmd = [
            '"{}"'.format(self._EXE),  # must wrap in quotes so that spaces don't mess things up
            '--install-source',
            'Anaconda-Navigator',
        ]
        logger.debug(' '.join(cmd))
        worker = wm.create_process_worker(cmd)
        return worker

    def _application_data(self):
        """Get app data from microsoft rest api."""
        data = {}
        url = (
            'https://update.code.visualstudio.com/api/update'
            '/{}/stable/version'.format(self._SUBDIR)
        )
        if url:
            data = self._download_api.get_url(
                url=url,
                as_json=True,
                non_blocking=False,
            )
        return data

    def _find_linux_install_dir(self):
        INST_DIR = None
        exe = os.path.join('/snap', 'bin', 'code')
        if os.path.lexists(exe):
            INST_DIR = '/snap'

        for distro in self.distro_map.keys():
            _distro_regex = ".*{}/([^ ]*)".format(distro)
            m = re.match(_distro_regex, self._conda_api.user_agent)
            if m:
                DISTRO_NAME = distro
                DISTRO_VER = m.group(1)
                break

        if DISTRO_NAME in ['ubuntu', 'debian']:
            _pkg_type = 'deb'
        else:
            _pkg_type = 'rpm'

        _os_arch = 'x64' if OS_64_BIT else 'ia32'
        SUBDIR = 'linux-{}-{}'.format(_pkg_type, _os_arch)
        INST_EXT = _pkg_type
        if not INST_DIR:
            INST_DIR = '/usr/share/{}'.format(self.filename)

        return SUBDIR, INST_EXT, INST_DIR

    def install(self, password=None):
        dummy_worker = DummyWorker()

        # On windows there is a User setup file also
        url = self.app_data.get('url')

        is_opensuse = 'opensuse' in self._DISTRO_NAME
        is_deb = self._DISTRO_NAME in ['ubuntu', 'debian']
        is_rpm = self._DISTRO_NAME in ['centos', 'rhel', 'fedora']
        wm = self._process_api

        if not url:
            self.trigger_finished_error(
                'Connectivity Error',
                'Please check your internet connection is working.',
                dummy_worker,
            )
            return dummy_worker

        if LINUX:
            if password is None:  # This should not happen, but just in case!
                self.trigger_finished_error(
                    'Password Error',
                    'Please try again and provide the correct credentials.',
                    dummy_worker,
                )
                return dummy_worker
            elif password == '':  # The install process was cancelled
                return dummy_worker

        def _download_finished(worker, output, error):
            """Download callback."""
            logger.debug('Finished App download')

            if error:
                dummy_worker.sig_finished.emit(dummy_worker, output, error)
                return

            if MAC:
                try:
                    os.makedirs(self._INST_DIR)
                except OSError as e:
                    if e.errno != errno.EEXIST:
                        logger.error(e)

                logger.debug('Decompressing app application')

                # Unzip using Mac defalut command/application
                command = [
                    '/usr/bin/unzip',
                    '-qo',
                    self._INSTFILE,
                    '-d',
                    self._INST_DIR,
                ]

                worker = wm.create_process_worker(command)
                worker.sig_partial.connect(dummy_worker.sig_partial)
                worker.sig_partial.emit(
                    dummy_worker,
                    {'message': 'Uncompressing file...'},
                    None,
                )
                worker.sig_finished.connect(_install_extensions)
                worker.start()
            elif WIN:
                # Run windows installer silently
                # When quotes are used  with START the first param is the title
                # that is why we add an empty string and then the actual
                # executable after the /WAIT. The quotes are for users with
                # spaces
                command = [
                    'START',
                    '/WAIT',
                    '""',
                    '"{}"'.format(self._INSTFILE),
                    '/VERYSILENT',
                    '/MERGETASKS=!runcode',
                    '/SUPPRESSMSGBOXES',
                    '/NORESTART',
                    '/LOG="{}"'.format(self.log_path(delete=True)),
                    '/DIR="{0}\\"'.format(self._INST_DIR),
                ]

                # Create temp batch file and run that
                cmd = u' '.join(command)  # The u'... is important on py27!
                logger.debug(cmd)
                bat_path = os.path.join(
                    CONF_PATH,
                    'temp',
                    'app-install.bat',
                )

                base_temp_path = os.path.dirname(bat_path)
                if not os.path.isdir(base_temp_path):
                    os.makedirs(base_temp_path)

                with io.open(bat_path, 'w') as f:
                    f.write(cmd)

                worker = wm.create_process_worker([bat_path])
                worker.sig_partial.connect(dummy_worker.sig_partial)
                worker.sig_finished.connect(_install_extensions)
                worker.start()

            elif LINUX:
                # See: https://code.visualstudio.com/docs/setup/linux
                if LINUX_DEB and is_deb:
                    cmd = ['sudo', '-kS', 'dpkg', '-i', self._INSTFILE]
                    worker = wm.create_process_worker(cmd)
                    worker.sig_partial.connect(dummy_worker.sig_partial)
                    worker.sig_finished.connect(_install_deb_dependencies)
                    worker.start()
                    stdin = to_binary_string(password + '\n')
                    worker.write(stdin)
                elif LINUX_RPM and is_rpm:
                    # Add key
                    cmd = [
                        'sudo', '-kS', 'rpm', '--import',
                        self.rpm_asc_file_url
                    ]
                    worker = wm.create_process_worker(cmd)
                    worker.sig_partial.connect(dummy_worker.sig_partial)
                    worker.sig_finished.connect(_install_rpm_repodata)
                    worker.start()
                    stdin = to_binary_string(password + '\n')
                    worker.write(stdin)
                else:
                    dummy_worker.sig_finished.emit(dummy_worker, None, None)

        def _install_deb_dependencies(worker, output, error):
            cmd = ['sudo', '-kS', 'apt-get', 'install', '-f']
            worker = wm.create_process_worker(cmd)
            worker.sig_partial.connect(dummy_worker.sig_partial)
            worker.sig_finished.connect(_install_extensions)
            worker.sig_partial.emit(
                dummy_worker,
                {'message': 'Installing dependencies...'},
                None,
            )
            worker.start()
            stdin = to_binary_string(password + '\n')
            worker.write(stdin)

        def _install_rpm_repodata(worker, output, error):
            logger.debug('install rpm repodata')
            distro = 'opensuse' if is_opensuse else 'fedora'
            cmd = [
                'sudo',
                '-kS',
                sys.executable,
                scripts.INSTALL_SCRIPT,
                distro,
            ]
            logger.debug(' '.join(cmd))
            worker = wm.create_process_worker(cmd)
            worker.sig_partial.connect(dummy_worker.sig_partial)
            worker.sig_finished.connect(_update_rpm_manager)
            worker.sig_partial.emit(
                dummy_worker,
                {'message': 'Installing repodata...'},
                None,
            )
            worker.start()
            stdin = to_binary_string(password + '\n')
            worker.write(stdin)

        def _update_rpm_manager(worker, output, error):
            logger.debug('update rpm manager')
            if is_opensuse:
                cmd = ['sudo', '-kS', 'zypper', 'refresh']
            elif LINUX_DNF:
                cmd = ['dnf', 'check-update']
            else:
                cmd = ['yum', 'check-update']
            logger.debug(' '.join(cmd))
            worker = wm.create_process_worker(cmd)
            worker.sig_partial.connect(dummy_worker.sig_partial)
            worker.sig_finished.connect(_install_rpm_package)
            worker.sig_partial.emit(
                dummy_worker,
                {'message': 'Updating manager...'},
                None,
            )
            worker.start()

            if is_opensuse:
                stdin = to_binary_string(password + '\n')
                worker.write(stdin)

        def _install_rpm_package(worker, output, error):
            logger.debug('install rpm package')
            if is_opensuse:
                cmd = [
                    'sudo', '-kS', 'zypper', '--non-interactive', 'install',
                    'code'
                ]
            elif LINUX_DNF:
                cmd = ['sudo', '-kS', 'dnf', '--assumeyes', 'install', 'code']
            else:
                cmd = ['sudo', '-kS', 'yum', '--assumeyes', 'install', 'code']

            logger.debug(' '.join(cmd))
            worker = wm.create_process_worker(cmd)
            worker.sig_partial.connect(dummy_worker.sig_partial)
            worker.sig_finished.connect(_install_extensions)
            worker.sig_partial.emit(
                dummy_worker,
                {'message': 'Installing rpm package...'},
                None,
            )
            worker.start()
            stdin = to_binary_string(password + '\n')
            worker.write(stdin)

        def _install_extensions(worker, output, error):
            """Install app extensions as part of install process."""
            logger.debug('install extensions')
            error = error.lower()
            check_in = 'error' in error
            check_not_in = 'password' not in error or 'sudo' not in error
            if error and check_in and check_not_in:
                dummy_worker.sig_finished.emit(dummy_worker, output, error)
                return

            worker = self.install_extensions()
            worker.sig_partial.emit(
                dummy_worker,
                {'message': 'Installing python extensions...'},
                None,
            )
            worker.sig_finished.connect(_send_telemetry)
            worker.start()

        def _send_telemetry(worker, output, error):
            """Send app telemetry as part of install process."""
            logger.debug('send telemetry')
            check_in = 'error' in error
            check_not_in = 'password' not in error or 'sudo' not in error
            if error and check_in and check_not_in:
                dummy_worker.sig_finished.emit(dummy_worker, output, error)
                return

            worker = self.send_telemetry()
            worker.sig_partial.emit(
                dummy_worker,
                {'message': 'Updating app data...'},
                None,
            )
            worker.sig_finished.connect(_installation_finished)
            worker.start()

        def _installation_finished(worker, output, error):
            # Check the log!
            logger.debug('Finished app installation')
            dummy_worker.sig_finished.emit(dummy_worker, output, error)

        # Download file
        worker = self._download_api.download(url, path=self._INSTFILE)
        worker.sig_partial.connect(dummy_worker.sig_partial)
        worker.sig_finished.connect(_download_finished)

        return dummy_worker

    def win_uninstaller(self):
        dats = []
        unins = []
        names = set()
        INST_FOLDER = self._INST_DIR
        for item in os.listdir(INST_FOLDER):
            path = os.path.join(INST_FOLDER, item).lower()
            item = item.lower()
            if os.path.isfile(path):
                parts = item.split('.')
                if parts:
                    basename = parts[0]
                else:
                    basename = item

                if item.startswith('unins') and item.endswith('.dat'):
                    dats.append(item)
                    names.add(basename)
                elif item.startswith('unins') and item.endswith('.exe'):
                    unins.append(item)
                    names.add(basename)
        for name in sorted(names):
            if name + '.dat' in dats and name + '.exe' in unins:
                # print(name + '.exe')
                break
        return name + '.exe' or 'unins000.exe'

    def remove(self, password=None):
        logger.debug('Removing app')
        wm = self._process_api
        dummy_worker = DummyWorker()
        locations = []
        uninstall_cmd = []
        is_opensuse = 'opensuse' in self._DISTRO_NAME
        is_deb = self._DISTRO_NAME in ['ubuntu', 'debian']
        is_rpm = self._DISTRO_NAME in ['centos', 'rhel', 'fedora']

        if MAC:
            locations = [self._INST_DIR]
        elif WIN:
            locations = [self._INST_DIR]
            uninstaller_path = os.path.join(
                self._INST_DIR,
                self.win_uninstaller(),
            )

            if ' ' in uninstaller_path:
                uninstaller_path = '"' + uninstaller_path + '"'
            command = [uninstaller_path]

            # Create temp batch file and run that
            cmd = ' '.join(command)
            bat_path = os.path.join(CONF_PATH, 'temp', 'app-uninstall.bat')
            base_temp_path = os.path.dirname(bat_path)

            if not os.path.isdir(base_temp_path):
                os.makedirs(base_temp_path)

            mode = 'w' if PY3 else 'wb'
            with io.open(bat_path, mode) as f:
                f.write(cmd)

            uninstall_cmd = [bat_path]

        elif LINUX:
            if LINUX_DEB and is_deb:
                uninstall_cmd = [
                    'sudo',
                    '-kS',
                    'apt-get',
                    '--yes',
                    'remove',
                    'code',
                ]
            elif LINUX_RPM and is_rpm:
                if is_opensuse:
                    uninstall_cmd = [
                        'sudo',
                        '-kS',
                        'zypper',
                        '--non-interactive',
                        'remove',
                        'code',
                    ]
                elif LINUX_DNF:
                    uninstall_cmd = [
                        'sudo',
                        '-kS',
                        'dnf',
                        '--assumeyes',
                        'remove',
                        'code',
                    ]
                else:
                    uninstall_cmd = [
                        'sudo',
                        '-kS',
                        'yum',
                        '--assumeyes',
                        'remove',
                        'code',
                    ]

        def _remove_locations(locations, dummy_worker):
            """Remove location."""
            for location in locations:
                if os.path.isdir(location):
                    logger.debug('Removing location: {}'.format(location))
                    try:
                        shutil.rmtree(location)
                    except Exception as e:
                        logger.debug(e)

            # Wait and check the folder has been indeed removed
            dummy_worker.sig_finished.emit(dummy_worker, {}, None)
            return locations

        def _uninstall_finished(worker, output, error):
            """Remove location callback."""
            new_worker = wm.create_python_worker(
                _remove_locations,
                locations,
                dummy_worker,
            )
            new_worker.sig_partial.connect(dummy_worker.sig_partial)
            new_worker.sig_finished.connect(_finished)
            new_worker.start()

        def _finished(worker, output, error):
            # Check uninstall log!
            log_data = self.log_data()
            if error or not log_data['successful']:
                out = {'error': error}
            else:
                out = output
            dummy_worker.sig_finished.emit(dummy_worker, out, None)

        if uninstall_cmd:
            logger.debug(' '.join(uninstall_cmd))
            worker = wm.create_process_worker(uninstall_cmd)
            worker.sig_partial.connect(dummy_worker.sig_partial)
            worker.sig_finished.connect(_uninstall_finished)
            worker.start()

            if LINUX:
                if password:
                    stdin = to_binary_string(password + '\n')
                    worker.write(stdin)
                else:
                    _uninstall_finished(dummy_worker, None, None)
        else:
            _uninstall_finished(dummy_worker, None, None)

        return dummy_worker

    def update_config(self, prefix):
        logger.debug('Update app config to use prefix {}'.format(prefix))
        try:
            _config = os.path.join(
                CONF_PATH,
                'Code',
                'User',
                'settings.json',
            )
            _config_dir = os.path.dirname(_config)

            try:
                if not os.path.isdir(_config_dir):
                    os.makedirs(_config_dir)
            except Exception as e:
                logger.error(e)

            config_update = {'python.pythonPath': prefix}

            if os.path.isfile(_config):
                try:
                    with io.open(_config, 'r', encoding='utf-8') as f:
                        data = f.read()

                    self.create_config_backup(data)

                    config_data = json.loads(data)
                    for key, val in config_update.items():
                        config_data[key] = val
                except Exception:
                    # If there is any error, don't overwrite app config
                    return False
            else:
                config_data = config_update.copy()

            mode = 'w' if PY3 else 'wb'
            with io.open(_config, mode) as f:
                json.dump(
                    config_data,
                    f,
                    sort_keys=True,
                    indent=4,
                )
        except Exception as e:
            logger.error(e)
            return False

        return True

    def create_config_backup(self, data):
        """
        Create a backup copy of the app configuration file `data`.

        Leave only the last 10 backups.
        """
        now = datetime.datetime.now()
        date = now.strftime('%Y%m%d%H%M%S')
        _config_dir = os.path.join(
            CONF_PATH,
            'Code',
            'User',
        )
        _config_bck = os.path.join(
            _config_dir,
            'bck.{date}.navigator.settings.json'.format(date=date),
        )

        # Make the backup
        logger.debug(
            'Creating backup app config file: {}'
            ''.format(_config_bck)
        )
        with io.open(_config_bck, 'w', encoding='utf-8') as f_handle:
            f_handle.write(data)

        # Only keep the latests 10 backups
        files = os.listdir(_config_dir)
        fpaths = [
            os.path.join(_config_dir, f) for f in files
            if f.startswith('bck.') and f.endswith('.navigator.settings.json')
        ]

        fpaths_remove = list(sorted(fpaths, reverse=True))[10:]
        for fpath in fpaths_remove:
            try:
                os.remove(fpath)
            except Exception:
                pass



import os

import xmltodict

from anaconda_navigator.static import images
from anaconda_navigator.utils.logs import logger
from anaconda_navigator.config import MAC, WIN
from .base import BaseApp


class PyCharmApp(BaseApp):
    def __init__(self, **kwargs):
        # minimum supported linux distro versions
        distro_map = {
            'rhel': '6',
            'sles': '11',
            'centos': '6',
            'debian': '7',
            'fedora': '20',
            'suse': '42.1',
            'ubuntu': '14.04'
        }

        super(PyCharmApp, self).__init__(
            app_name='pycharm',
            filename='pycharm',
            windows_ext='.exe',
            windows_folder_name='JetBrains',
            mac_name='PyCharm.app',
            display_name='PyCharm',
            description=('Full-featured Python IDE by JetBrains.  Supports '
                         'code completion, linting, debugging, and '
                         'domain-specific enhancements for web development '
                         'and data science.'),
            image_path=images.PYCHARM_ICON_1024_PATH,
            needs_license=False,
            non_conda=True,
            distro_map=distro_map,
            **kwargs
        )

        self._appdata = None
        self.app_data = self._application_data()

    def _find_linux_install_dir(self):
        if os.path.isdir('/snap/bin') and any(_.startswith('pycharm') for _ in os.listdir('/snap/bin')):
            install_dir = '/snap'
        else:
            install_dir = '/opt'
        return None, None, install_dir

    @property
    def executable(self):
        launch_str = None
        if MAC:
            # pycharm will be available as a .app bundle
            dirs = [name for name in os.listdir(self._INST_DIR)
                    if os.path.isdir(os.path.join(self._INST_DIR, name)) if name.startswith("PyCharm")]
            found_pycharm_app = dirs[0] if dirs else None
            launch_str = 'open -a "{}"'.format(found_pycharm_app.rstrip('.app')) if found_pycharm_app else None
        elif WIN:
            if os.path.isdir(self._INST_DIR):
                for root, dirs, files in os.walk(self._INST_DIR):
                    for file in files:
                        if file.lower().startswith('pycharm') and file.lower().endswith(".exe"):
                            return '"{}"'.format(os.path.join(root, file))
        else:
            # snaps on Ubuntu have this naming.  They are ideally on PATH.
            for edition in ('professional', 'community', 'educational'):
                exe = os.path.join(self._INST_DIR, 'bin', 'pycharm-{}'.format(edition))
                if os.path.lexists(exe):
                    launch_str = exe
            # this is the other method listed on JetBrains' site
            # https://www.jetbrains.com/help/pycharm/installation-guide.html#snap-install-tar
            if not launch_str and os.path.isdir(self._INST_DIR):
                pycharm_dirs = [_ for _ in os.listdir(self._INST_DIR) if os.path.isdir(_) and _.startswith('pycharm')]
                for pc_dir in pycharm_dirs:
                    for root, dirs, files in os.walk(pc_dir):
                        for file in files:
                            if file.lower() == 'pycharm.sh':
                                return 'sh "{}"'.format(os.path.join(root, file))
        return launch_str

    def _application_data(self):
        if self._appdata:
            return self._appdata
        url = "https://www.jetbrains.com/updates/updates.xml"
        data = self._download_api.get_url(
            url=url,
            as_json=False,
            non_blocking=False,
        )
        rel = {}
        if data:
            data = xmltodict.parse(data)
            p = [_ for _ in data['products']['product'] if _['@name'] == 'PyCharm'][0]
            rel = [_ for _ in p['channel'] if _['@status'] == 'release'][0]
        self._appdata = rel
        return rel

    @property
    def versions(self):
        versions = set()
        for build in self._application_data().get('build', []):
            versions.add(build['@version'])
            # for k in list(build.keys()):
            #     if k.startswith("@"):
            #         build[k[1:]] = build[k]
            #         del build[k]
        return sorted(list(versions))

    def _pycharm_version(self):
        version = None
        # Version info is in a file called build.txt.  On mac, it is at
        # Contents/Resources/build.txt
        # from vlan: 193 means 2019.3 (EAP or release or 2019.3.x updates), 201 means 2020.1, etc.
        return ""

    @property
    def version(self):
        """Return the current installed version or the highest version."""
        return self._pycharm_version() or self.versions[-1] if self.versions else None

    def create_config_backup(self, data):
        """
        Create a backup copy of the app configuration file `data`.

        Leave only the last 10 backups.
        """
        # no effect for pycharm
        pass

    def update_config(self, prefix):
        """Update app python interpreter user config."""
        # no effect for pycharm
        pass

    def install_extensions(self):
        """Install app extensions."""
        # no effect for pycharm
        wm = self._process_api
        worker = wm.create_process_worker(['python', '-V'])
        return worker

    def install(self, password=None):
        """Install app."""
        raise NotImplementedError

    def send_telemtry(self):
        """Send app telemetry."""
        raise NotImplementedError

    def win_uninstaller(self):
        """Check the right uninstaller file on windows."""
        raise NotImplementedError

    def remove(self, password=None):
        """Remove app files from computer or run uninstaller."""
        raise NotImplementedError



"""
The MIT License (MIT)

Copyright (c) 2012-2014 Alexander Turkin
Copyright (c) 2014 William Hallatt
Copyright (c) 2015 Jacob Dawid
Copyright (c) 2016 Luca Weiss

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# Standard library imports
import math

# Third party imports
from qtpy.QtCore import QRect, Qt, QTimer
from qtpy.QtGui import QColor, QPainter
from qtpy.QtWidgets import QWidget


class QtWaitingSpinner(QWidget):
    """"""

    def __init__(
        self,
        parent,
        centerOnParent=True,
        disableParentWhenSpinning=False,
        modality=Qt.NonModal
    ):
        """"""
        super(QtWaitingSpinner, self).__init__(parent=parent)

        self._centerOnParent = centerOnParent
        self._disableParentWhenSpinning = disableParentWhenSpinning

        # WAS IN initialize()
        self._color = QColor(Qt.black)
        self._roundness = 100.0
        self._minimumTrailOpacity = 3.14159265358979323846
        self._trailFadePercentage = 80.0
        self._revolutionsPerSecond = 1.57079632679489661923
        self._numberOfLines = 20
        self._lineLength = 10
        self._lineWidth = 2
        self._innerRadius = 10
        self._currentCounter = 0
        self._isSpinning = False

        self._timer = QTimer(self)
        self._timer.timeout.connect(self.rotate)
        self.updateSize()
        self.updateTimer()
        self.hide()
        # END initialize()

        self.setWindowModality(modality)
        self.setAttribute(Qt.WA_TranslucentBackground)

    def paintEvent(self, QPaintEvent):
        """"""
        self.updatePosition()
        painter = QPainter(self)
        painter.fillRect(self.rect(), Qt.transparent)
        painter.setRenderHint(QPainter.Antialiasing, True)

        if self._currentCounter >= self._numberOfLines:
            self._currentCounter = 0

        painter.setPen(Qt.NoPen)
        for i in range(0, self._numberOfLines):
            painter.save()
            painter.translate(
                self._innerRadius + self._lineLength,
                self._innerRadius + self._lineLength
            )
            rotateAngle = float(360 * i) / float(self._numberOfLines)
            painter.rotate(rotateAngle)
            painter.translate(self._innerRadius, 0)
            distance = self.lineCountDistanceFromPrimary(
                i, self._currentCounter, self._numberOfLines
            )
            color = self.currentLineColor(
                distance, self._numberOfLines, self._trailFadePercentage,
                self._minimumTrailOpacity, self._color
            )
            painter.setBrush(color)
            painter.drawRoundedRect(
                QRect(
                    0, -self._lineWidth / 2, self._lineLength, self._lineWidth
                ), self._roundness, self._roundness, Qt.RelativeSize
            )
            painter.restore()

    def start(self):
        """"""
        self.updatePosition()
        self._isSpinning = True
        self.show()

        if self.parentWidget and self._disableParentWhenSpinning:
            self.parentWidget().setEnabled(False)

        if not self._timer.isActive():
            self._timer.start()
            self._currentCounter = 0

    def stop(self):
        """"""
        self._isSpinning = False
        self.hide()

        if self.parentWidget() and self._disableParentWhenSpinning:
            self.parentWidget().setEnabled(True)

        if self._timer.isActive():
            self._timer.stop()
            self._currentCounter = 0

    def setNumberOfLines(self, lines):
        """"""
        self._numberOfLines = lines
        self._currentCounter = 0
        self.updateTimer()

    def setLineLength(self, length):
        """"""
        self._lineLength = length
        self.updateSize()

    def setLineWidth(self, width):
        """"""
        self._lineWidth = width
        self.updateSize()

    def setInnerRadius(self, radius):
        """"""
        self._innerRadius = radius
        self.updateSize()

    def color(self):
        """"""
        return self._color

    def roundness(self):
        """"""
        return self._roundness

    def minimumTrailOpacity(self):
        """"""
        return self._minimumTrailOpacity

    def trailFadePercentage(self):
        """"""
        return self._trailFadePercentage

    def revolutionsPersSecond(self):
        """"""
        return self._revolutionsPerSecond

    def numberOfLines(self):
        """"""
        return self._numberOfLines

    def lineLength(self):
        """"""
        return self._lineLength

    def lineWidth(self):
        """"""
        return self._lineWidth

    def innerRadius(self):
        """"""
        return self._innerRadius

    def isSpinning(self):
        """"""
        return self._isSpinning

    def setRoundness(self, roundness):
        """"""
        self._roundness = max(0.0, min(100.0, roundness))

    def setColor(self, color=Qt.black):
        """"""
        self._color = QColor(color)

    def setRevolutionsPerSecond(self, revolutionsPerSecond):
        """"""
        self._revolutionsPerSecond = revolutionsPerSecond
        self.updateTimer()

    def setTrailFadePercentage(self, trail):
        """"""
        self._trailFadePercentage = trail

    def setMinimumTrailOpacity(self, minimumTrailOpacity):
        """"""
        self._minimumTrailOpacity = minimumTrailOpacity

    def rotate(self):
        """"""
        self._currentCounter += 1
        if self._currentCounter >= self._numberOfLines:
            self._currentCounter = 0
        self.update()

    def updateSize(self):
        """"""
        size = (self._innerRadius + self._lineLength) * 2
        self.setFixedSize(size, size)

    def updateTimer(self):
        """"""
        self._timer.setInterval(
            1000 / (self._numberOfLines * self._revolutionsPerSecond)
        )

    def updatePosition(self):
        """"""
        if self.parentWidget() and self._centerOnParent:
            self.move(
                self.parentWidget().width() / 2 - self.width() / 2,
                self.parentWidget().height() / 2 - self.height() / 2
            )

    def lineCountDistanceFromPrimary(self, current, primary, totalNrOfLines):
        """"""
        distance = primary - current
        if distance < 0:
            distance += totalNrOfLines
        return distance

    def currentLineColor(
        self, countDistance, totalNrOfLines, trailFadePerc, minOpacity,
        colorinput
    ):
        """"""
        color = QColor(colorinput)
        if countDistance == 0:
            return color
        minAlphaF = minOpacity / 100.0
        distanceThreshold = int(
            math.ceil((totalNrOfLines - 1) * trailFadePerc / 100.0)
        )
        if countDistance > distanceThreshold:
            color.setAlphaF(minAlphaF)
        else:
            alphaDiff = color.alphaF() - minAlphaF
            gradient = alphaDiff / float(distanceThreshold + 1)
            resultAlpha = color.alphaF() - gradient * countDistance
            # If alpha is out of bounds, clip it.
            resultAlpha = min(1.0, max(0.0, resultAlpha))
            color.setAlphaF(resultAlpha)
        return color


class NavigatorSpinner(QtWaitingSpinner):
    """"""

    def __init__(self, parent, total_width=20):
        """"""
        super(NavigatorSpinner, self).__init__(parent=parent)

        # Based on anaconda logo proportions
        r0 = 4.68
        w0 = 2.32
        radius = total_width / (((2 * w0) / r0) + 2)
        ring = radius * w0 / r0

        # Set custom configuration
        self.setInnerRadius(radius)
        self.setLineLength(ring)
        self.setLineWidth(ring)
        self.setNumberOfLines(90)
        self.setColor(QColor('#43B02A'))
        self.setRevolutionsPerSecond(0.50)
        self.setRoundness(100.0)


if __name__ == '__main__':
    from anaconda_navigator.utils.qthelpers import qapplication
    app = qapplication()
    widget = NavigatorSpinner(None)
    widget.start()
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2020 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Tool path utilties."""

import os
import sys

def get_pyexec(prefix):
    """Provides the full path to python executable"""
    rslt = ''
    if sys.platform == "win32":
        rslt += os.sep.join([prefix, 'python.exe'])
    else:
        rslt += os.sep.join([prefix, 'bin', 'python'])
    return rslt

def get_pyscript(prefix, name):
    """Provides the OS dependent path in bin/Scripts for python script tool"""
    rslt = ''
    if sys.platform == "win32":
        rslt += os.sep.join([prefix, 'Scripts', name + '-script.py'])
    else:
        rslt += os.sep.join([prefix, 'bin', name])
    return rslt

